import datetime

import pytest
from django.urls import reverse

from api.models import (
    Keys,
    Minions,
    SaltReturns,
    MinionsCustomFields,
    Schedule,
    UserSettings,
)


@pytest.mark.django_db()
def test_index(admin_client):
    response = admin_client.get(reverse("index"))
    assert response.status_code == 200


@pytest.mark.django_db()
def test_keys_list(key, admin_client, jwt):
    response = admin_client.get("/api/keys/", **jwt)
    assert len(response.json()) == Keys.objects.count()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_keys_refresh(key, admin_client, jwt):
    response = admin_client.post("/api/keys/refresh/", {}, **jwt)
    assert response.json()["result"] == "refreshed"
    assert response.status_code == 200


@pytest.mark.django_db()
def test_keys_status(key, admin_client, jwt):
    response = admin_client.get("/api/keys/keys_status/", **jwt)
    assert "accepted" in response.json()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_keys_manage(key, admin_client, jwt):
    response = admin_client.post(
        "/api/keys/manage_keys/", {"target": "master", "action": "reject"}, **jwt
    )
    assert "result" in response.json()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_list(minion, admin_client, jwt):
    response = admin_client.get("/api/minions/", **jwt)
    assert response.json()[0]["minion_id"] == "2e220fd40bc5"
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_delete(minion, admin_client, jwt):
    response = admin_client.delete("/api/minions/{}/".format(minion.minion_id), **jwt)
    assert response.status_code == 204


@pytest.mark.django_db()
def test_minions_refresh(minion, admin_client, jwt):
    response = admin_client.post(
        "/api/minions/refresh_minions/", {"minion_id": minion.minion_id}, **jwt
    )
    assert "result" in response.json()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_conformity(minion, admin_client, jwt):
    response = admin_client.get("/api/minions/conformity/", **jwt)
    assert "HIGHSTATE" in response.json()["name"]
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_conformity_change(minion, highstate_diff, admin_client, jwt):
    response = admin_client.get("/api/minions/conformity/", **jwt)
    assert "HIGHSTATE" in response.json()["name"]
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_conformity_detail(minion, highstate, admin_client, jwt):
    highstate()
    response = admin_client.get(
        "/api/minions/{}/conformity_detail/".format(minion.minion_id), **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_conformity_detail_highstate(minion, highstate_diff, admin_client, jwt):
    response = admin_client.get(
        "/api/minions/{}/conformity_detail/".format(minion.minion_id), **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_conformity_detail_empty(minion_master, admin_client, jwt):
    response = admin_client.get(
        "/api/minions/{}/conformity_detail/".format(minion_master.minion_id), **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minions_conformity_render(minion, highstate, minion_master, admin_client, jwt):
    highstate()
    response = admin_client.get("/api/conformity/render/", **jwt)
    assert "name" in response.json()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_minion_field_delete(admin_client, jwt):
    response = admin_client.post(
        "/api/minionsfields/",
        {"name": "highstate", "function": "state.show_highstate", "value": "{}"},
        **jwt
    )
    assert response.status_code == 201

    response = admin_client.post(
        "/api/minionsfields/delete_field/", {"name": "highstate"}, **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_schedules_list(schedule, admin_client, jwt):
    response = admin_client.get("/api/schedules/", **jwt)
    assert response.json()[0]["minion"] == "master"
    assert response.json()[0]["name"] == "job2"
    assert response.status_code == 200


@pytest.mark.django_db()
def test_schedules_refresh(admin_client, jwt):
    response = admin_client.post("/api/schedules/refresh/", {}, **jwt)
    assert "result" in response.json()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_schedules_manage(admin_client, jwt):
    response = admin_client.post(
        "/api/run/",
        {
            "command": "salt master schedule.add job2 function='test.ping' seconds=3600",
            "raw": True,
        },
        **jwt
    )
    assert response.status_code == 200
    response = admin_client.post("/api/schedules/refresh/", {}, **jwt)
    assert response.status_code == 200
    response = admin_client.post(
        "/api/schedules/manage/",
        {"action": "delete", "name": "job2", "minion": "master"},
        **jwt
    )
    assert "result" in response.json()
    assert response.status_code == 200


def test_users_list(admin_client, admin_user, jwt):
    response = admin_client.get("/api/users/", **jwt)
    assert response.json()[0]["id"] == admin_user.id
    assert response.status_code == 200


def test_users_refresh_token(admin_client, admin_user, jwt):
    current_token = admin_user.user_settings.token
    response = admin_client.post(
        "/api/users/{}/manage_token/".format(admin_user.id), {"action": "renew"}, **jwt
    )
    assert response.status_code == 200
    assert UserSettings.objects.get(user=admin_user).token != current_token


def test_users_revoke_token(admin_client, admin_user, jwt):
    response = admin_client.post(
        "/api/users/{}/manage_token/".format(admin_user.id), {"action": "revoke"}, **jwt
    )
    assert response.status_code == 200
    assert UserSettings.objects.get(user=admin_user).token == "REVOKED"


def test_jobs_graph(admin_client, jwt):
    response = admin_client.get("/api/jobs/graph?period=2", **jwt)
    assert "labels" in response.json()
    assert response.status_code == 200


def test_jobs_graph_highstate(admin_client, jwt):
    response = admin_client.get("/api/jobs/graph?period=2&fun=highstate", **jwt)
    assert "labels" in response.json()
    assert response.status_code == 200


def test_graph_other_filter(admin_client, highstate, dummy_state, jwt):
    highstate()
    response = admin_client.get("/api/jobs/graph?period=0&fun=other", **jwt)
    assert response.status_code == 200
    assert response.json()["series"][0][0] > 1


def test_stats(admin_client, jwt):
    response = admin_client.get("/api/stats/", **jwt)
    for i in ["jobs", "events", "schedules"]:
        assert i in response.json()
    assert response.status_code == 200


def test_get_events(admin_client):
    response = admin_client.get("/api/event_stream/")
    assert response.status_code == 200


@pytest.mark.django_db()
def test_run_scheduled_cron(admin_client, jwt):
    response = admin_client.post(
        "/api/run/",
        {
            "raw": "true",
            "command": "salt * test.ping",
            "schedule": "true",
            "schedule_type": "cron",
            "cron": "* * * * *",
        },
        **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_run_scheduled_once(admin_client, jwt):
    response = admin_client.post(
        "/api/run/",
        {
            "raw": "true",
            "command": "salt * test.ping",
            "schedule": "{}".format(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            ),
            "schedule_type": "once",
        },
        **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_get_functions(admin_client, jwt):
    response = admin_client.get("/api/functions/", **jwt)
    assert response.status_code == 200


@pytest.mark.django_db()
def test_wheel_raw(admin_client, jwt):
    response = admin_client.post(
        "/api/run/",
        {"command": "salt --client=runner pillar.show_top", "raw": True},
        **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_search_minion(admin_client, dummy_state, minion_master, jwt):
    response = admin_client.get("/api/search/?q=master", **jwt)
    assert response.json()["query"] == "master"
    assert response.json()["minions"][0]["minion_id"] == "master"


@pytest.mark.django_db()
def test_search_job(admin_client, dummy_state, dummy_jid, minion_master, jwt):
    response = admin_client.get("/api/search/?q=20190507190955945844", **jwt)
    assert response.json()["query"] == "20190507190955945844"
    assert response.json()["jobs"][0]["jid"] == dummy_state.jid



import pytest
from api.models import SaltReturns, SaltEvents


@pytest.mark.django_db()
def test_salt_returns_list(admin_client, jwt):
    response = admin_client.get("/api/jobs/", **jwt)
    assert len(response.json()) == SaltReturns.objects.count()
    assert response.status_code == 200


@pytest.mark.django_db()
def test_salt_returns_filters(jid, highstate, admin_client, jwt):
    highstate()
    response = admin_client.get("/api/jobs/filters/", **jwt)
    assert "2e220fd40bc5" in response.json()["minions"]
    assert response.status_code == 200


@pytest.mark.django_db()
def test_salt_returns_list_filtered(jid, highstate, admin_client, jwt):
    highstate()
    response = admin_client.get("/api/jobs/?limit=1&target[]=2e220fd40bc5", **jwt)
    assert len(response.json()) <= 1
    assert response.status_code == 200


@pytest.mark.django_db()
def test_salt_returns_job(jid, highstate, admin_client, jwt):
    highstate()
    assert SaltReturns.objects.get(jid="20190429180928455927", id="2e220fd40bc5")
    response = admin_client.get("/api/jobs/20190429180928455927/2e220fd40bc5/", **jwt)
    assert response.json()["jid"] == "20190429180928455927"
    assert response.status_code == 200


@pytest.mark.django_db()
def test_salt_returns_list(admin_client, jwt):
    response = admin_client.get("/api/events/", **jwt)
    assert len(response.json()) <= 200
    assert response.status_code == 200


@pytest.mark.django_db()
def test_salt_returns_job_rendered(jid, highstate, admin_client, jwt):
    highstate()
    assert SaltReturns.objects.get(jid="20190429180928455927", id="2e220fd40bc5")
    response = admin_client.get(
        "/api/jobs/20190429180928455927/2e220fd40bc5/rendered_state/", **jwt
    )
    assert response.status_code == 200



import pytest
from django.urls import reverse

from api.models import Keys, Minions, Functions, Schedule, SaltReturns


def test_demo_default_admin(admin_client):
    """
    Make sure dummy user is available.
    """
    response = admin_client.get(reverse("index"), follow=True)
    assert response.status_code == 200


@pytest.mark.django_db
def test_default_perms(client, admin_user):
    """
    Default permissions are added to salt, and stored in user_settings.
    """
    # Set perms.
    username = "admin"
    password = "password"
    client.login(username=username, password=password)
    client.get(reverse("index"), follow=True)
    assert hasattr(admin_user.user_settings, "salt_permissions")


@pytest.mark.django_db()
def test_add_keys(admin_client, jwt):
    """
    user_settings are working as expected, and /wheel behave correctly.
    """
    response = admin_client.post(
        "/api/keys/manage_keys/", {"action": "accept", "target": "*"}, **jwt
    )
    assert response.status_code == 200
    assert hasattr(response, "json")
    resp = response.json()
    assert resp == {"result": "accept on *: done"}


@pytest.mark.django_db()
def test_add_minions(admin_client, jwt):
    assert Minions.objects.count() == 0
    response = admin_client.post("/api/minions/refresh_minions/", **jwt)
    assert response.status_code == 200
    assert Minions.objects.count() > 0


@pytest.mark.django_db()
def test_run_highstate(admin_client, jwt):
    for minion in Minions.objects.all():
        assert minion.conformity() is False
    response = admin_client.post(
        "/api/run/", {"raw": "true", "command": "salt * state.apply"}, **jwt
    )
    assert response.status_code == 200
    assert SaltReturns.objects.filter(fun="state.apply")
    for minion in Minions.objects.all():
        assert minion.conformity() is True


@pytest.mark.django_db()
def test_runner(admin_client, jwt):
    response = admin_client.post(
        "/api/run/",
        {"raw": "true", "command": "salt --client=runner jobs.active"},
        **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_wheel(admin_client, jwt):
    response = admin_client.post(
        "/api/run/",
        {"raw": "true", "command": "salt --client=wheel key.name_match master"},
        **jwt
    )
    assert response.status_code == 200
    jobs = SaltReturns.objects.all()


@pytest.mark.django_db()
def test_run_conformity_state(admin_client, jwt):
    response = admin_client.post(
        "/api/run/", {"raw": "true", "command": "salt * grains.item os"}, **jwt
    )
    response = admin_client.post(
        "/api/run/", {"command": "salt * grains.item saltversion", "raw": True}, **jwt
    )
    for minion in Minions.objects.all():
        assert minion.custom_conformity("grains.item", "os")
        assert minion.custom_conformity("grains.item", "saltversion")


@pytest.mark.django_db()
def test_refresh_keys(admin_client, jwt):
    keys_attr = ["minion_id", "pub", "status"]
    current_keys = list(Keys.objects.filter().values_list(*keys_attr))
    response = admin_client.post("/api/keys/refresh/", {}, **jwt)
    assert response.status_code == 200
    updated_keys = list(Keys.objects.filter().values_list(*keys_attr))
    assert current_keys == updated_keys


@pytest.mark.django_db()
def test_init_db(admin_client, jwt):
    assert Functions.objects.count() == 0
    response = admin_client.post("/api/settings/initdb", {"target": "master"}, **jwt)
    assert response.status_code == 200
    for funct_type in ["local", "runner", "wheel"]:
        assert Functions.objects.filter(type=funct_type).count() > 0


@pytest.mark.django_db()
def test_add_minion_field(admin_client, jwt):
    response = admin_client.post(
        "/api/minionsfields/",
        {"name": "highstate", "function": "state.show_highstate", "value": "{}"},
        **jwt
    )
    assert response.status_code == 201


@pytest.mark.django_db()
def test_change_notifs(admin_client, admin_user, jwt):
    """
    Default permissions are added to salt, and stored in user_settings.
    """
    notifs_defaults = {
        "notifs_created": False,
        "notifs_published": False,
        "notifs_returned": True,
        "notifs_event": False,
    }
    for status, value in notifs_defaults.items():
        assert getattr(admin_user.user_settings, status) == value
    response = admin_client.patch(
        "/api/userssettings/{}/".format(admin_user.id),
        {"notifs_created": "true"},
        content_type="application/json",
        **jwt
    )
    assert response.status_code == 200


@pytest.mark.django_db()
def test_schedules(admin_client, run_sql, jwt):
    assert Schedule.objects.count() == 0
    response = admin_client.post(
        "/api/run/",
        {
            "command": "salt master schedule.add job1 function='test.ping' seconds=3600",
            "raw": True,
        },
        **jwt
    )
    assert response.status_code == 200
    response = admin_client.post("/api/schedules/refresh/", {}, **jwt)
    assert response.status_code == 200
    assert len(Schedule.objects.all()) > 0
    # run_sql("TRUNCATE TABLE `salt_returns`")



"""
Django settings for Alcali project.

"""
import errno
import os

# from datetime import timedelta
# If there's our env var, it means that env file was loaded somehow(docker).
from os.path import join
from pathlib import Path

from dotenv import load_dotenv

DB_BACKEND = os.environ.get("DB_BACKEND")
if not DB_BACKEND:
    # Load env file
    ENV_PATH = os.environ.get("ENV_PATH", os.getcwd())
    if not ENV_PATH:
        raise FileNotFoundError("ENV_PATH is not set")
    dotenv_path = join(ENV_PATH, ".env")
    env_file = Path(dotenv_path)
    if not env_file.exists():
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), env_file)
    load_dotenv(dotenv_path)

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ["SECRET_KEY"]

# SECURITY WARNING: don't run with debug turned on in production!
from distutils.util import strtobool

DJANGO_DEBUG = os.environ.get("DJANGO_DEBUG", None)
if DJANGO_DEBUG:
    try:
        # lower('y', 'yes', 't', 'true', 'on', '1')
        DJANGO_DEBUG = strtobool(DJANGO_DEBUG)
    # None, empty, bool...
    except (AttributeError, TypeError, ValueError):
        DJANGO_DEBUG = False
else:
    DJANGO_DEBUG = False
DEBUG = DJANGO_DEBUG

ALLOWED_HOSTS = [os.environ.get("ALLOWED_HOSTS", "127.0.0.1")]

# Application definition

INSTALLED_APPS = [
    "api.apps.ApiConfig",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "whitenoise.runserver_nostatic",
    "django.contrib.staticfiles",
    "rest_framework",
    "corsheaders",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",  # TODO: check priority
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_currentuser.middleware.ThreadLocalUserMiddleware",
]

ROOT_URLCONF = "config.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "dist")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

WSGI_APPLICATION = "config.wsgi.application"

# Database
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.{}".format(os.environ["DB_BACKEND"]),
        "NAME": os.environ.get("DB_NAME"),
        "USER": os.environ.get("DB_USER"),
        "PASSWORD": os.environ.get("DB_PASS"),
        "HOST": os.environ.get("DB_HOST"),
        "PORT": os.environ.get("DB_PORT"),
    }
}

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

# TIME_ZONE = "Europe/Paris"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = "/static/"

# Place static in the same location as webpack build files
STATIC_ROOT = os.path.join(BASE_DIR, "dist", "static")
STATICFILES_DIRS = []

##########
# STATIC #
##########

STATICFILES_STORAGE = "whitenoise.storage.CompressedManifestStaticFilesStorage"

REST_FRAMEWORK = {
    "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.IsAuthenticated",),
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework_simplejwt.authentication.JWTAuthentication"
    ],
}
# CORS_URLS_REGEX = r'^/api/.*$'
# CORS_ORIGIN_WHITELIST = [
#     "http://127.0.0.1:8001"
# ]

# SIMPLE_JWT = {
#     'ACCESS_TOKEN_LIFETIME': timedelta(days=1),
#     'REFRESH_TOKEN_LIFETIME': timedelta(days=1)}



import datetime
import json
from collections import Counter, OrderedDict

from ansi2html import Ansi2HTMLConverter
from django.contrib.auth.models import User
from django.db.models import Q
from django.http import (
    HttpResponse,
    JsonResponse,
    StreamingHttpResponse,
    HttpResponseRedirect,
)
from django.views.decorators.cache import never_cache
from django.views.generic import TemplateView
from rest_framework import viewsets
from rest_framework.decorators import (
    action,
    api_view,
    renderer_classes,
    permission_classes,
)
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework_simplejwt.views import TokenObtainPairView

from api.backend.netapi import (
    refresh_minion,
    manage_key,
    get_events,
    init_db,
    refresh_schedules,
    run_raw,
    get_keys,
    manage_schedules,
)
from api.models import SaltReturns, Keys, Minions, SaltEvents, Schedule, Conformity
from api.models import UserSettings, MinionsCustomFields, Functions
from api.permissions import IsLoggedInUserOrAdmin, IsAdminUser
from api.renderer import StreamingRenderer
from api.serializers import (
    ConformitySerializer,
    UsersSerializer,
    UserSettingsSerializer,
    MinionsCustomFieldsSerializer,
    FunctionsSerializer,
    ScheduleSerializer,
    MyTokenObtainPairSerializer,
    SaltReturnsSerializer,
)
from api.serializers import KeysSerializer, MinionsSerializer
from api.utils import graph_data, render_conformity, RawCommand
from api.utils.output import highstate_output, nested_output

# Serve Vue Application
index_view = never_cache(TemplateView.as_view(template_name="index.html"))


class KeysViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Keys.objects.all()
    serializer_class = KeysSerializer

    @action(methods=["POST"], detail=False)
    def refresh(self, request):
        get_keys(refresh=True)
        return Response({"result": "refreshed"})

    @action(detail=False)
    def keys_status(self, request):
        keys_status = list(Keys.objects.values_list("status", flat=True))
        keys_status = dict(Counter(keys_status))
        od = OrderedDict()
        for status in ["accepted", "rejected", "denied", "unaccepted"]:
            if status not in keys_status:
                od[status] = 0
            else:
                od[status] = keys_status[status]
        return Response(od)

    @action(methods=["post"], detail=False)
    def manage_keys(self, request):
        kwargs = {}
        key = request.data.get("target")
        key_action = request.data.get("action")
        if key_action == "accept":
            kwargs = {"include_rejected": True, "include_denied": True}
        elif key_action == "reject":
            kwargs = {"include_accepted": True, "include_denied": True}
        response = manage_key(key_action, key, kwargs)
        return Response({"result": "{} on {}: done".format(key_action, key)})


class MinionsViewSet(viewsets.ModelViewSet):
    queryset = Minions.objects.all()
    serializer_class = MinionsSerializer
    lookup_field = "minion_id"
    lookup_value_regex = "[0-9a-zA-Z.]+"

    @action(detail=False, methods=["post"])
    def refresh_minions(self, request):
        if request.POST.get("minion_id"):
            minion_id = request.POST.get("minion_id")
            refresh_minion(minion_id)
            return Response({"result": "refreshed {}".format(minion_id)})

        accepted_minions = Keys.objects.filter(status="accepted").values_list(
            "minion_id", flat=True
        )
        for minion in accepted_minions:
            refresh_minion(minion)
        return Response({"refreshed": [i for i in accepted_minions]})

    @action(detail=False)
    def conformity(self, request):
        highstate_conformity = {"conform": 0, "conflict": 0, "unknown": 0}
        for minion in Minions.objects.all():
            if minion.conformity() is True:
                highstate_conformity["conform"] += 1
            elif minion.conformity() is False:
                highstate_conformity["conflict"] += 1
            else:
                highstate_conformity["unknown"] += 1

        conformity_name, conformity_data, _ = render_conformity()
        conformity_name.insert(0, "HIGHSTATE")
        conformity_data.insert(0, highstate_conformity)
        return Response({"name": conformity_name, "data": conformity_data})

    @action(detail=True)
    def conformity_detail(self, request, minion_id):
        minion = self.get_object()

        # Get conformity data.
        _, _, custom_conformity = render_conformity(minion.minion_id)
        custom_conformity = (
            custom_conformity[minion.minion_id] if custom_conformity else None
        )
        minion_conformity = minion.conformity()

        # Convert states to html.
        conv = Ansi2HTMLConverter(inline=False, scheme="xterm")
        last_highstate = minion.last_highstate()
        succeeded, unchanged, failed = {}, {}, {}

        if last_highstate:
            last_highstate = last_highstate.loaded_ret()["return"]
            # Sls error
            if isinstance(last_highstate, list):
                failed = {"error": last_highstate[0]}
            else:
                for state in last_highstate:
                    state_name = state.split("_|-")[1]
                    formatted = highstate_output.output(
                        {minion.minion_id: {state: last_highstate[state]}},
                        summary=False,
                    )
                    if last_highstate[state]["result"] is True:
                        succeeded[state_name] = conv.convert(
                            formatted, ensure_trailing_newline=True
                        )
                    elif last_highstate[state]["result"] is None:
                        unchanged[state_name] = conv.convert(
                            formatted, ensure_trailing_newline=True
                        )
                    else:
                        failed[state_name] = conv.convert(
                            formatted, ensure_trailing_newline=True
                        )
        return Response(
            {
                "custom_conformity": custom_conformity,
                "conformity": minion_conformity,
                "succeeded": succeeded,
                "unchanged": unchanged,
                "failed": failed,
            }
        )


class MinionsCustomFieldsViewSet(viewsets.ModelViewSet):
    queryset = MinionsCustomFields.objects.all()
    serializer_class = MinionsCustomFieldsSerializer

    def perform_create(self, serializer):
        for minion in Minions.objects.all():
            serializer.save(minion=minion)

    @action(methods=["post"], detail=False)
    def delete_field(self, request):
        field = request.data.get("name")
        MinionsCustomFields.objects.filter(name=field).delete()
        return Response({"result": "{} field deleted".format(field)})


class ConformityViewSet(viewsets.ModelViewSet):
    queryset = Conformity.objects.all()
    serializer_class = ConformitySerializer
    lookup_value_regex = "[0-9a-zA-Z.]+"

    @action(detail=False)
    def render(self, request):
        conformity_name = [i.name for i in Conformity.objects.all()]
        default_columns = [
            {"text": "Minion id", "value": "minion_id"},
            {"text": "Last Highstate", "value": "last_highstate"},
            {"text": "Highstate Conformity", "value": "conformity"},
            {"text": "Succeeded", "value": "succeeded"},
            {"text": "Unchanged", "value": "unchanged"},
            {"text": "Failed", "value": "failed"},
        ]
        for conformity in conformity_name:
            default_columns.append({"text": conformity, "value": conformity})

        # Get conformity data.
        _, _, rendered_conformity = render_conformity()
        # Compute number of succeeded, unchanged and failed states.
        conformity_data = []
        minions = Minions.objects.all()
        for minion in minions:
            succeeded, unchanged, failed = 0, 0, 0
            last_highstate = minion.last_highstate()
            if last_highstate:
                last_highstate_date = last_highstate.alter_time
                last_highstate = last_highstate.loaded_ret()["return"]
                # Sls error
                if isinstance(last_highstate, list):
                    succeeded, unchanged, failed = None, None, 1
                else:
                    for state in last_highstate:
                        if last_highstate[state]["result"] is True:
                            succeeded += 1
                        elif last_highstate[state]["result"] is None:
                            unchanged += 1
                        else:
                            failed += 1
            else:
                last_highstate_date, succeeded, unchanged, failed = (
                    None,
                    None,
                    None,
                    None,
                )

            default_conformity = {
                "minion_id": minion.minion_id,
                "last_highstate": last_highstate_date,
                "conformity": minion.conformity(),
                "succeeded": succeeded,
                "unchanged": unchanged,
                "failed": failed,
            }
            # Add custom conformity values to datatable data.
            if rendered_conformity:
                for conformity in rendered_conformity[minion.minion_id]:
                    default_conformity.update(conformity)
            conformity_data.append(default_conformity)
        return Response({"name": default_columns, "data": conformity_data})


class FunctionsViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Functions.objects.all()
    serializer_class = FunctionsSerializer


class ScheduleViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Schedule.objects.all()
    serializer_class = ScheduleSerializer

    def list(self, request, *args, **kwargs):
        # Datatable.
        ret = []
        for schedule in Schedule.objects.values():
            for k, v in json.loads(schedule["job"]).items():
                schedule[k] = v
            del schedule["job"]
            ret.append(schedule)
        return Response(ret)

    @action(methods=["POST"], detail=False)
    def refresh(self, request):
        refresh_schedules()
        return Response({"result": "refreshed"})

    @action(methods=["POST"], detail=False)
    def manage(self, request):
        action = request.data.get("action")
        minion = request.data.get("minion")
        name = request.data.get("name")
        manage_schedules(action, name, minion)
        return Response(
            {"result": "schedule " + name + " on " + minion + " " + action + "d"}
        )


class UsersViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UsersSerializer

    def get_permissions(self):
        permission_classes = []
        if self.action == "create":
            permission_classes = [AllowAny]
        elif (
            self.action == "retrieve"
            or self.action == "update"
            or self.action == "partial_update"
        ):
            permission_classes = [IsLoggedInUserOrAdmin]
        elif self.action == "list" or self.action == "destroy":
            permission_classes = [IsAdminUser]
        return [permission() for permission in permission_classes]

    @action(methods=["POST"], detail=True)
    def manage_token(self, request, pk):
        user = self.get_object()
        action = request.data.get("action")
        if action == "renew":
            user.user_settings.generate_token()
        elif action == "revoke":
            user.user_settings.token = "REVOKED"
            user.user_settings.save()
        return Response({"result": "{} successful".format(action)})


class UserSettingsViewSet(viewsets.ModelViewSet):
    queryset = UserSettings.objects.all()
    serializer_class = UserSettingsSerializer


@api_view(["GET"])
def jobs_graph(request):
    id = request.query_params.get("id", None)
    params = {
        "period": int(request.query_params.get("period", 7)),
        "fun": request.query_params.get("fun", "all"),
    }
    if id:
        params.update(id=id)
    days, count, error_count = graph_data(**params)
    return Response({"labels": days, "series": [count, error_count]})


@api_view(["POST"])
def parse_modules(request):
    if request.data.get("target"):
        init_db(request.data.get("target"))
        return Response({"result": "modules updated"})


@api_view(["GET"])
def search(request):
    if request.query_params.get("q", None):
        query = request.query_params.get("q")
        # First try to match minions.
        minion_results = []
        return_results = []
        minion_query = Minions.objects.filter(minion_id__icontains=query)
        return_query = SaltReturns.objects.filter(
            Q(jid__icontains=query) | Q(fun__icontains=query)
        )[:200]
        if minion_query:
            for res in minion_query:
                minion = MinionsSerializer(res)
                minion_results.append(minion.data)
        if return_query:
            for res in return_query:
                job = SaltReturnsSerializer(res)
                return_results.append(job.data)
        return Response(
            {"minions": minion_results, "jobs": return_results, "query": query}
        )

        # Return to referer
    return HttpResponseRedirect(request.META.get("HTTP_REFERER"))


@api_view(["GET"])
def stats(request):
    # Status widget.
    jobs_nb = SaltReturns.objects.count()
    events_nb = SaltEvents.objects.count()
    schedules_nb = Schedule.objects.count()
    return Response({"jobs": jobs_nb, "events": events_nb, "schedules": schedules_nb})


@api_view(["GET"])
@permission_classes([AllowAny])
@renderer_classes([StreamingRenderer])
def event_stream(request):
    # Web socket.
    response = StreamingHttpResponse(
        get_events(), status=200, content_type="text/event-stream"
    )
    response["Cache-Control"] = "no-cache"
    return response


@api_view(["POST"])
def run(request):
    if request.POST.get("raw"):
        command = RawCommand(request.POST.get("command"))
        parsed_command = command.parse()
        # Schedules.
        if request.POST.get("schedule_type"):
            schedule_type = request.POST.get("schedule_type")
            schedule_name = request.POST.get(
                "schedule_name", datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            )
            schedule_parsed = [
                {
                    "client": "local",
                    "batch": None,
                    "tgt_type": parsed_command[0]["tgt_type"],
                    "tgt": parsed_command[0]["tgt"],
                    "fun": "schedule.add",
                    "arg": [
                        schedule_name,
                        "function={}".format(parsed_command[0]["fun"]),
                        "job_args={}".format(parsed_command[0]["arg"]),
                    ],
                }
            ]
            if schedule_type == "once":
                schedule_date = request.POST.get("schedule")
                schedule_parsed[0]["arg"].append("once={}".format(schedule_date))
                schedule_parsed[0]["arg"].append("once_fmt=%Y-%m-%d %H:%M:%S")
            else:
                cron = request.POST.get("cron")
                schedule_parsed[0]["arg"].append("cron={}".format(cron))
            ret = run_raw(schedule_parsed)
            formatted = nested_output.output(ret)
            conv = Ansi2HTMLConverter(inline=False, scheme="xterm")
            html = conv.convert(formatted, ensure_trailing_newline=True)
            return HttpResponse(html)

        ret = run_raw(parsed_command)
        formatted = "\n"
        if (
            parsed_command[0]["fun"] in ["state.apply", "state.highstate"]
            and parsed_command[0]["client"] != "local_async"
        ):
            for state, out in ret.items():
                minion_ret = highstate_output.output({state: out})
                formatted += minion_ret + "\n\n"
        else:
            for state, out in ret.items():
                minion_ret = nested_output.output({state: out})
                formatted += minion_ret + "\n\n"
        if request.POST.get("cli"):
            return JsonResponse({"results": formatted})
        conv = Ansi2HTMLConverter(inline=False, scheme="xterm")
        html = conv.convert(formatted, ensure_trailing_newline=True)
        return HttpResponse(html)


class MyTokenObtainPairView(TokenObtainPairView):
    serializer_class = MyTokenObtainPairSerializer



from rest_framework import permissions


class IsLoggedInUserOrAdmin(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj == request.user or request.user.is_staff


class IsAdminUser(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.user and request.user.is_staff

    def has_object_permission(self, request, view, obj):
        return request.user and request.user.is_staff



import os
import json
import urllib3

from django.contrib.auth.models import User
from pepper import Pepper
from django_currentuser.middleware import get_current_user

from ..utils.input import RawCommand
from ..models import Minions, Functions, MinionsCustomFields, Keys, Schedule

urllib3.disable_warnings()

url = os.environ.get("SALT_URL", "https://127.0.0.1:8080")


def api_connect():
    # TODO fix this!
    user = get_current_user()
    api = Pepper(url, ignore_ssl_errors=True)
    login_ret = api.login(
        str(user.username),
        user.user_settings.token,
        os.environ.get("SALT_AUTH", "alcali"),
    )
    user.user_settings.salt_permissions = json.dumps(login_ret["perms"])
    user.save()
    return api

    # except (PepperException, ConnectionRefusedError, URLError) as e:
    #     print("Can't connect to {url}: {e}".format(url=url, e=e))


def get_keys(refresh=False):
    if refresh:
        # Salt return to minion status.
        minion_status = {
            "minions_rejected": "rejected",
            "minions_denied": "denied",
            "minions_pre": "unaccepted",
            "minions": "accepted",
        }

        api = api_connect()
        api_ret = api.wheel("key.list_all")["return"][0]["data"]["return"]

        Keys.objects.all().delete()
        for key, value in minion_status.items():
            for minion in api_ret[key]:
                finger_ret = api.wheel("key.finger", match=minion, hash_type="sha256")[
                    "return"
                ][0]["data"]["return"][key]
                obj, created = Keys.objects.update_or_create(
                    minion_id=minion,
                    defaults={"status": value, "pub": finger_ret[minion]},
                )
                if created:
                    pass
                    # LOG CREATED

    return Keys.objects.all()


def refresh_minion(minion_id):
    api = api_connect()
    grain = api.local(minion_id, "grains.items")
    grain = grain["return"][0]
    # TODO: return smt useful, better error mgmt.
    if grain.get(minion_id):
        pillar = api.local(minion_id, "pillar.items")
        pillar = pillar["return"][0]
        Minions.objects.update_or_create(
            minion_id=minion_id,
            defaults={
                "grain": json.dumps(grain[minion_id]),
                "pillar": json.dumps(pillar[minion_id]),
            },
        )
        minion_fields = MinionsCustomFields.objects.values(
            "name", "function"
        ).distinct()
        for field in minion_fields:
            command = RawCommand("salt {} {}".format(minion_id, field["function"]))
            custom_field_return = run_raw(command.parse())
            MinionsCustomFields.objects.update_or_create(
                name=field["name"],
                function=field["function"],
                minion=Minions.objects.get(minion_id=minion_id),
                defaults={"value": json.dumps(custom_field_return[minion_id])},
            )


def run_raw(load):
    api = api_connect()
    api_ret = api.low(load)
    api_ret = api_ret["return"][0]
    return api_ret


def get_events():
    api = Pepper(url, ignore_ssl_errors=True)
    # TODO: find a way.
    user = User.objects.first()
    api.login(
        str(user.username),
        user.user_settings.token,
        os.environ.get("SALT_AUTH", "alcali"),
    )
    response = api.req_stream("/events")
    return response


def init_db(target):
    api = api_connect()
    # Modules.
    modules_func = api.local(target, "sys.list_functions")
    modules_func = modules_func["return"][0][target]

    modules_doc = api.local(target, "sys.doc")

    for func in modules_func:
        desc = modules_doc["return"][0][target][func]

        Functions.objects.update_or_create(
            name=func, type="local", description=desc or ""
        )
    # Runner.
    # TODO: Factorize.
    runner_func = api.local(target, "sys.list_runner_functions")
    runner_func = runner_func["return"][0][target]

    runner_doc = api.local(target, "sys.runner_doc")

    for func in runner_func:
        desc = runner_doc["return"][0][target][func]

        Functions.objects.update_or_create(
            name=func, type="runner", description=desc or ""
        )
    wheel_docs = api.runner("doc.wheel")
    wheel_docs = wheel_docs["return"][0]
    for fun, doc in wheel_docs.items():
        Functions.objects.update_or_create(
            name=fun, type="wheel", description=doc or ""
        )
    return {"something": "useful"}


def manage_key(action, target, kwargs):
    api = api_connect()
    response = api.wheel("key.{}".format(action), match=target, **kwargs)
    return response


def refresh_schedules(minion=None):
    minion = minion or "*"
    api = api_connect()
    api_ret = api.local(minion, "schedule.list", kwarg={"return_yaml": False})
    for minion_id in api_ret["return"][0]:
        # TODO: error mgmt
        minion_jobs = api_ret["return"][0][minion_id]
        Schedule.objects.filter(minion=minion_id).delete()
        for job_name in minion_jobs:
            if job_name != "schedule":
                Schedule.objects.create(
                    minion=minion_id,
                    name=job_name,
                    job=json.dumps(minion_jobs[job_name]),
                )
    return api_ret["return"][0]


def manage_schedules(action, name, minion):
    api = api_connect()
    api_ret = api.local(minion, "schedule.{}".format(action), arg=name)
    for target in api_ret["return"][0]:
        # If action was successful.
        if api_ret["return"][0][target]["result"]:
            if "delete" in action:
                Schedule.objects.filter(minion=minion, name=name).delete()
                # TODO: disable, enable
            # else:
            #     try:
            #         sched = Schedule.objects.filter(minion=minion, name=name).get()
            #     except Schedule.DoesNotExist:
            #         refresh_schedules(minion)
            #         try:
            #             sched = Schedule.objects.filter(minion=minion, name=name).get()
            #         except Schedule.DoesNotExist:
            #             return False
            #     loaded_job = sched.loaded_job()
            #     if "enable" in action:
            #         loaded_job["enabled"] = True
            #     elif "disable" in action:
            #         loaded_job["enabled"] = False
            #     sched.job = json.dumps(loaded_job)
            #     sched.save()



# -*- coding: utf-8 -*-
"""
Outputter for displaying results of state runs
==============================================

The return data from the Highstate command is a standard data structure
which is parsed by the highstate outputter to deliver a clean and readable
set of information about the HighState run on minions.

Two configurations can be set to modify the highstate outputter. These values
can be set in the master config to change the output of the ``salt`` command or
set in the minion config to change the output of the ``salt-call`` command.

state_verbose:
    By default `state_verbose` is set to `True`, setting this to `False` will
    instruct the highstate outputter to omit displaying anything in green, this
    means that nothing with a result of True and no changes will not be printed
state_output:
    The highstate outputter has five output modes, ``full``, ``terse``,
    ``mixed``, ``changes`` and ``filter``.

    * The default is set to ``full``, which will display many lines of detailed
      information for each executed chunk.
    * If ``terse`` is used, then the output is greatly simplified and shown in
      only one line.
    * If ``mixed`` is used, then terse output will be used unless a state
      failed, in which case full output will be used.
    * If ``changes`` is used, then terse output will be used if there was no
      error and no changes, otherwise full output will be used.
    * If ``filter`` is used, then either or both of two different filters can be
      used: ``exclude`` or ``terse``.
      These can be set as such from the command line, or in the Salt config as
      `state_output_exclude` or `state_output_terse`, respectively. The values to
      exclude must be a comma-separated list of `True`, `False` and/or `None`.
      Because of parsing nuances, if only one of these is used, it must still
      contain a comma. For instance: `exclude=True,`.
state_tabular:
    If `state_output` uses the terse output, set this to `True` for an aligned
    output format.  If you wish to use a custom format, this can be set to a
    string.

Example output:

.. code-block:: text

    myminion:
    ----------
              ID: test.ping
        Function: module.run
          Result: True
         Comment: Module function test.ping executed
         Changes:
                  ----------
                  ret:
                      True

    Summary
    ------------
    Succeeded: 1
    Failed:    0
    ------------
    Total:     0
"""

# Import python libs
from __future__ import absolute_import
import pprint
import textwrap

from . import nested_output


def output(data, summary=True):
    """
    The HighState Outputter is only meant to be used with the state.highstate
    function, or a function that returns highstate return data.
    """
    for host, hostdata in data.items():
        return _format_host(host, hostdata, summary)[0]


colors = {
    "GREEN": "\033[1;32;40m",
    "CYAN": "\033[1;36;40m",
    "LIGHT_RED": "\033[1;31;40m",
    "YELLOW": "\033[1;33;40m",
    "RED": "\033[1;31;40m",
    "ENDC": "\033[0;0m",
}


def _format_host(host, data, summary):
    tabular = False
    rcounts = {}
    rdurations = []
    hcolor = colors["GREEN"]
    hstrs = []
    nchanges = 0

    if isinstance(data, int) or isinstance(data, str):
        # Data in this format is from saltmod.function,
        # so it is always a 'change'
        nchanges = 1
        hstrs.append((u"{0}    {1}{2[ENDC]}".format(hcolor, data, colors)))
        hcolor = colors["CYAN"]  # Print the minion name in cyan
    if isinstance(data, list):
        # Errors have been detected, list them in RED!
        hcolor = colors["LIGHT_RED"]
        hstrs.append(
            (u"    {0}Data failed to compile:{1[ENDC]}".format(hcolor, colors))
        )
        for err in data:
            hstrs.append(
                (u"{0}----------\n    {1}{2[ENDC]}".format(hcolor, err, colors))
            )
    if isinstance(data, dict):
        # Verify that the needed data is present
        for tname, info in data.items():
            if isinstance(info, dict) and "__run_num__" not in info:
                err = (
                    u"The State execution failed to record the order "
                    "in which all states were executed. The state "
                    "return missing data is:"
                )
                hstrs.insert(0, pprint.pformat(info))
                hstrs.insert(0, err)
        # Everything rendered as it should display the output
        for tname in sorted(data, key=lambda k: data[k].get("__run_num__", 0)):
            ret = data[tname]
            # Increment result counts
            rcounts.setdefault(ret["result"], 0)
            rcounts[ret["result"]] += 1
            rduration = ret.get("duration", 0)
            try:
                rdurations.append(float(rduration))
            except ValueError:
                rduration, _, _ = rduration.partition(" ms")
                try:
                    rdurations.append(float(rduration))
                except ValueError:
                    pass
                    # log.error('Cannot parse a float from duration %s', ret.get('duration', 0))

            tcolor = colors["GREEN"]
            schanged, ctext = _format_changes(ret["changes"])
            nchanges += 1 if schanged else 0

            if schanged:
                tcolor = colors["CYAN"]
            if ret["result"] is False:
                hcolor = colors["RED"]
                tcolor = colors["RED"]
            if ret["result"] is None:
                hcolor = colors["YELLOW"]
                tcolor = colors["YELLOW"]
            comps = tname.split("_|-")
            state_lines = [
                u"{tcolor}----------{colors[ENDC]}",
                u"    {tcolor}      ID: {comps[1]}{colors[ENDC]}",
                u"    {tcolor}Function: {comps[0]}.{comps[3]}{colors[ENDC]}",
                u"    {tcolor}  Result: {ret[result]!s}{colors[ENDC]}",
                u"    {tcolor} Comment: {comment}{colors[ENDC]}",
                u"    {tcolor} Started: {ret[start_time]!s}{colors[ENDC]}",
                u"    {tcolor}Duration: {ret[duration]!s}{colors[ENDC]}",
            ]
            # This isn't the prettiest way of doing this, but it's readable.
            if comps[1] != comps[2]:
                state_lines.insert(3, u"    {tcolor}    Name: {comps[2]}{colors[ENDC]}")
            try:
                comment = ret["comment"]
                comment = comment.strip().replace(u"\n", u"\n" + u" " * 14)
            except AttributeError:  # Assume comment is a list
                try:
                    comment = ret["comment"].join(" ").replace(u"\n", u"\n" + u" " * 13)
                except AttributeError:
                    # Comment isn't a list either, just convert to string
                    comment = str(ret["comment"])
                    comment = comment.strip().replace(u"\n", u"\n" + u" " * 14)
            for detail in ["start_time", "duration"]:
                ret.setdefault(detail, u"")
            if ret["duration"] != "":
                ret["duration"] = u"{0} ms".format(ret["duration"])
            svars = {
                "tcolor": tcolor,
                "comps": comps,
                "ret": ret,
                "comment": comment,
                # This nukes any trailing \n and indents the others.
                "colors": colors,
            }
            hstrs.extend([sline.format(**svars) for sline in state_lines])
            changes = u"     Changes:   " + ctext
            hstrs.append((u"{0}{1}{2[ENDC]}".format(tcolor, changes, colors)))

            if "warnings" in ret:
                rcounts.setdefault("warnings", 0)
                rcounts["warnings"] += 1
                wrapper = textwrap.TextWrapper(
                    width=80, initial_indent=u" " * 14, subsequent_indent=u" " * 14
                )
                hstrs.append(
                    u"   {colors[LIGHT_RED]} Warnings: {0}{colors[ENDC]}".format(
                        wrapper.fill("\n".join(ret["warnings"])).lstrip(), colors=colors
                    )
                )

        # Append result counts to end of output
        colorfmt = u"{0}{1}{2[ENDC]}"
        rlabel = {
            True: u"Succeeded",
            False: u"Failed",
            None: u"Not Run",
            "warnings": u"Warnings",
        }
        count_max_len = max([len(str(x)) for x in rcounts.items()] or [0])
        label_max_len = max([len(x) for x in rlabel.items()] or [0])
        line_max_len = label_max_len + count_max_len + 2  # +2 for ': '
        if summary:
            hstrs.append(
                colorfmt.format(
                    colors["CYAN"], u"\nSummary\n{0}".format("-" * line_max_len), colors
                )
            )

            def _counts(label, count):
                return u"{0}: {1:>{2}}".format(
                    label, count, line_max_len - (len(label) + 2)
                )

            # Successful states
            changestats = []
            if None in rcounts and rcounts.get(None, 0) > 0:
                # test=True states
                changestats.append(
                    colorfmt.format(
                        colors["YELLOW"],
                        u"unchanged={0}".format(rcounts.get(None, 0)),
                        colors,
                    )
                )
            if nchanges > 0:
                changestats.append(
                    colorfmt.format(
                        colors["GREEN"], u"changed={0}".format(nchanges), colors
                    )
                )
            if changestats:
                changestats = u" ({0})".format(", ".join(changestats))
            else:
                changestats = u""
            hstrs.append(
                colorfmt.format(
                    colors["GREEN"],
                    _counts(rlabel[True], rcounts.get(True, 0) + rcounts.get(None, 0)),
                    colors,
                )
                + changestats
            )

            # Failed states
            num_failed = rcounts.get(False, 0)
            hstrs.append(
                colorfmt.format(
                    colors["RED"] if num_failed else colors["CYAN"],
                    _counts(rlabel[False], num_failed),
                    colors,
                )
            )

            num_warnings = rcounts.get("warnings", 0)
            if num_warnings:
                hstrs.append(
                    colorfmt.format(
                        colors["LIGHT_RED"],
                        _counts(rlabel["warnings"], num_warnings),
                        colors,
                    )
                )

            totals = u"{0}\nTotal states run: {1:>{2}}".format(
                "-" * line_max_len,
                sum(rcounts.values()) - rcounts.get("warnings", 0),
                line_max_len - 7,
            )
            hstrs.append(colorfmt.format(colors["CYAN"], totals, colors))
            sum_duration = sum(rdurations)
            duration_unit = "ms"
            # convert to seconds if duration is 1000ms or more
            if sum_duration > 999:
                sum_duration /= 1000
                duration_unit = "s"
            total_duration = "Total run time: {0} {1}".format(
                "{0:.3f}".format(sum_duration).rjust(line_max_len - 5), duration_unit
            )
            hstrs.append(colorfmt.format(colors["CYAN"], total_duration, colors))

    hstrs.insert(0, (u"{0}{1}:{2[ENDC]}".format(hcolor, host, colors)))
    return u"\n".join(hstrs), nchanges > 0


def _format_changes(changes):
    """
    Format the changes dict based on what the data is
    """

    if not changes:
        return False, u""

    if not isinstance(changes, dict):
        return True, u"Invalid Changes data: {0}".format(changes)

    ret = changes.get("ret")
    if ret is not None and changes.get("out") == "highstate":
        ctext = u""
        changed = False
        for host, hostdata in ret.items():
            s, c = _format_host(host, hostdata)
            ctext += u"\n" + u"\n".join((u" " * 14 + l) for l in s.splitlines())
            changed = changed or c
    else:
        changed = True
        ctext = nested_output.output(changes, nested_indent=14)
    return changed, ctext



from collections import Counter

# from difflib import SequenceMatcher
from datetime import timedelta

from django.utils import timezone
from django.db.models import Count, Q

from .input import RawCommand
from ..models import SaltReturns
from ..models import Minions, Conformity


# def auto_group(minion_list):
#     return SequenceMatcher(None, [i for i in minion_list]).find_longest_match([0])
#

# Find all occurrences of a key in nested python dictionaries and lists.
# def find_key(key, dictionary):
#     for k, v in dictionary.items():
#         if k == key:
#             yield v
#         elif isinstance(v, dict):
#             for result in find_key(key, v):
#                 yield result
#         elif isinstance(v, list):
#             for d in v:
#                 for result in find_key(key, d):
#                     yield result
#


def graph_data(period=7, **kwargs):
    """
    Retrieve graph data, count nb of jobs per days.
    :param period: nb of days
    :return: (days, count): list of days, nb of jobs.
    """
    period_frame = timezone.now().date() - timedelta(days=period)

    if kwargs.get("fun") == "highstate":
        kwargs.pop("fun")
        data = (
            SaltReturns.objects.filter(
                Q(fun="state.apply") | Q(fun="state.highstate"),
                alter_time__gt=period_frame,
                **kwargs
            )
            .extra({"day": "date(alter_time)"})
            .values("day")
            .annotate(count=Count("jid"))
            .annotate(error=Count("success", filter=Q(success="0")))
            .order_by("day")
        )

    elif kwargs.get("fun") == "other":
        kwargs.pop("fun")
        data = (
            SaltReturns.objects.filter(alter_time__gt=period_frame, **kwargs)
            .exclude(Q(fun="state.apply") | Q(fun="state.highstate"))
            .extra({"day": "date(alter_time)"})
            .values("day")
            .annotate(count=Count("jid"))
            .annotate(error=Count("success", filter=Q(success="0")))
            .order_by("day")
        )

    else:
        kwargs.pop("fun")
        data = (
            SaltReturns.objects.filter(alter_time__gt=period_frame, **kwargs)
            .extra({"day": "date(alter_time)"})
            .values("day")
            .annotate(count=Count("jid"))
            .annotate(error=Count("success", filter=Q(success="0")))
            .order_by("day")
        )

    days = []
    count = []
    error_count = []
    for res in data:
        days.append(str(res["day"]))
        count.append(res["count"])
        error_count.append(res["error"])
    for idx, date in enumerate(period_frame + timedelta(n) for n in range(period)):
        # Complete all lists with default data.
        if str(date) not in days:
            days.insert(idx, str(date))
            count.insert(idx, 0)
            error_count.insert(idx, 0)

    return days, count, error_count


def render_conformity(target=None):

    # First, a list of targets.
    if target:
        minions_all = [Minions.objects.get(minion_id=target)]
    else:
        minions_all = Minions.objects.all()

    # Get conformity fields.
    conformity_fields = Conformity.objects.values("name", "function")

    # Default.
    if not conformity_fields:
        return [], [], []

    # Create list of conformity names in UPPERCASE.
    conformity_names = [i["name"].upper() for i in conformity_fields]

    # Aggregated results structure.
    ret = []

    # Detailed structure.
    details = {minion.minion_id: [] for minion in minions_all}

    for field in conformity_fields:
        field_ret = []

        # Parse command.
        funct = field["function"]
        # Use fake client because we don't provide a target
        parser = RawCommand(funct, client="runner", inline=True)
        parsed = parser.parse()[0]

        for minion in minions_all:
            if "arg" in parsed:
                conformity_field = minion.custom_conformity(
                    parsed["fun"], *parsed["arg"]
                )
            else:
                conformity_field = minion.custom_conformity(parsed["fun"])

            # If the job return is a dict, get the value.
            if isinstance(conformity_field, dict):
                conformity_ret = list(conformity_field.values())[0]
                field_ret.append(conformity_ret)
                details[minion.minion_id].append({field["name"]: conformity_ret})
            # It should be a string.
            else:
                field_ret.append(conformity_field)
                details[minion.minion_id].append({field["name"]: conformity_field})
        ret.append(dict(Counter(field_ret)))

    return conformity_names, ret, details



import os
from io import StringIO

import pytest
from django.core.management import call_command, CommandError


@pytest.mark.django_db
def test_check():
    out = StringIO()
    call_command("check", stdout=out)
    assert "db:\tok" in out.getvalue()
    assert "env:\tok" in out.getvalue()


@pytest.mark.django_db
def test_check_env_fail():
    out = StringIO()
    salt_url = os.environ["SALT_URL"]
    del os.environ["SALT_URL"]
    call_command("check", stdout=out)
    assert "db:\tok" in out.getvalue()
    assert "SALT_URL" in out.getvalue()
    os.environ["SALT_URL"] = str(salt_url)


@pytest.mark.django_db
def test_manage_token(admin_user):
    current_token = admin_user.user_settings.token
    out = StringIO()
    call_command("manage_token", "admin", stdout=out)
    assert current_token in out.getvalue()


@pytest.mark.django_db
def test_manage_token_raises():
    out = StringIO()
    with pytest.raises(CommandError) as err:
        call_command("manage_token", "foo", stdout=out)
    assert "user foo does" in str(err.value)


@pytest.mark.django_db
def test_manage_token_reset(admin_user):
    current_token = admin_user.user_settings.token
    out = StringIO()
    call_command("manage_token", "admin", "-r", stdout=out)
    assert current_token not in out.getvalue()



from django.db.models.signals import pre_save, post_save

from django.dispatch import receiver
from django.contrib.auth.models import User

from api.models import UserSettings


@receiver(post_save, sender=User)
def create_user_settings(sender, instance, created, **kwargs):

    # Create related UserSettings when a new User is created.
    if created:
        UserSettings.objects.create(user=instance)


@receiver(post_save, sender=User)
def save_user_settings(sender, instance, **kwargs):

    # Save UserSettings when user is saved.
    instance.user_settings.save()



import time
import pytest
from django.contrib.auth.models import User

from api.models import SaltReturns, Jids, Keys, Minions, Schedule, Conformity


@pytest.fixture
def jwt(admin_client):
    res = admin_client.post(
        "/api/token/", data={"username": "admin", "password": "password"}
    )
    token = res.data["access"]
    return {"HTTP_AUTHORIZATION": "Bearer {}".format(token)}


@pytest.fixture
def jid():
    Jids.objects.create(
        jid="20190429180928455927",
        load='{"tgt_type": "glob", "jid": "20190429180928455927", '
        '"tgt": "2e220fd40bc5", "cmd": "publish", "ret": "", '
        '"user": "admin", "key": '
        '"N6TS4++wIWU/7jOQsSYO+E42VbFc/b/j3nD9GC24Oo08jF'
        '/bY49vgksEHUMmRNzLXTuzsdiYLwM=", "kwargs": {"token": '
        '"3ee91b11a58f8265c26db179dd61244b12db375d9855d4b2487033f7'
        '540710ca", "client": "local", "batch": null}, "fun": '
        '"state.apply", "arg": []}',
    )


@pytest.fixture
def key():
    Keys.objects.create(
        minion_id="2e220fd40bc5",
        pub="d5:20:87:6f:85:4c:51:52:9e:63:08:45:04:fd:61:d8:53:5b:c8"
        ":15:08:21:d3:6a:85:8e:27:0b:e8:d4:93:e9",
        status="accepted",
    )


@pytest.fixture
def highstate():
    def _highstate(minion="2e220fd40bc5"):
        date = time.strftime("%Y-%m-%d %H:%M:%S")
        ret = SaltReturns.objects.create(
            fun="state.apply",
            jid="20190429180928455927",
            return_field='{"pkg_|-common_packages_'
            '|-common_packages_|-installed": {'
            '"comment": "The following packages '
            "were "
            "installed/updated: strace, htop, "
            'vim", '
            '"name": "common_packages", '
            '"start_time": '
            '"18:09:30.691635", "result": true, '
            '"duration": 25031.334, '
            '"__run_num__": 0, '
            '"__sls__": "common.packages", '
            '"changes": '
            '{"vim-common": {"new": '
            '"2:8.0.0197-4+deb9u1", "old": ""}, '
            '"strace": {"new": "4.15-2", "old": '
            '""}, '
            '"libgpm2": {"new": "1.20.4-6.2+b1", '
            '"old": ""}, "vim": {"new": '
            '"2:8.0.0197-4+deb9u1", "old": ""}, '
            '"htop": {"new": "2.0.2-1", "old": '
            '""}, '
            '"xxd": {"new": '
            '"2:8.0.0197-4+deb9u1", '
            '"old": ""}, "vim-runtime": {"new": '
            '"2:8.0.0197-4+deb9u1", "old": ""}}, '
            '"__id__": "common_packages"}}',
            id="2e220fd40bc5",
            success=1,
            full_ret='{"fun_args": [], "jid": '
            '"20190429180928455927", "return": {'
            '"pkg_|-common_packages_|-common_packages_'
            '|-installed": {"comment": "The following '
            "packages were installed/updated: strace, "
            'htop, vim", "name": "common_packages", '
            '"start_time": "18:09:30.691635", '
            '"result": '
            'true, "duration": 25031.334, '
            '"__run_num__": '
            '0, "__sls__": "common.packages", '
            '"changes": { '
            '"vim-common": {"new": '
            '"2:8.0.0197-4+deb9u1", '
            '"old": ""}, "strace": {"new": "4.15-2", '
            '"old": ""}, "libgpm2": {"new": '
            '"1.20.4-6.2+b1", "old": ""}, "vim": {'
            '"new": '
            '"2:8.0.0197-4+deb9u1", "old": ""}, '
            '"htop": { '
            '"new": "2.0.2-1", "old": ""}, "xxd": {'
            '"new": '
            '"2:8.0.0197-4+deb9u1", "old": ""}, '
            '"vim-runtime": {"new": '
            '"2:8.0.0197-4+deb9u1", '
            '"old": ""}}, "__id__": '
            '"common_packages"}}, '
            '"retcode": 0, "success": true, '
            '"cmd": "_return", "_stamp": '
            '"2019-04-29T18:09:55.727933", '
            '"fun": "state.apply", "id": '
            '"2e220fd40bc5", '
            '"out": "highstate"}',
            alter_time=date,
        )
        return ret

    return _highstate


@pytest.fixture
def highstate_diff():
    ret = SaltReturns.objects.create(
        fun="state.apply",
        jid="20190507190955945843",
        return_field='{"pkg_|-common_packages_'
        '|-common_packages_|-installed": {'
        '"comment": "The following packages '
        "would "
        'be installed/updated: vim", "name": '
        '"common_packages", "start_time": '
        '"19:09:59.565546", "result": null, '
        '"duration": 412.241, "__run_num__": '
        "0, "
        '"__sls__": "common.packages", '
        '"changes": '
        '{}, "__id__": "common_packages"}, '
        '"pkg_|-install_alcali_'
        "|-install_alcali_ "
        '|-installed": {"comment": "All '
        "specified "
        'packages are already installed", '
        '"name": '
        '"curl", "start_time": '
        '"19:09:59.978059", '
        '"result": true, "duration": 34.526, '
        '"__run_num__": 1, "__sls__": '
        '"alcali", '
        '"changes": {}, "__id__": '
        '"install_alcali"}}',
        id="2e220fd40bc5",
        success=1,
        full_ret='{"fun_args": ["test=True"], "jid": '
        '"20190507190955945843", "return": {'
        '"pkg_|-common_packages_|-common_packages_'
        '|-installed": {"comment": "The following '
        "packages would be installed/updated: "
        'vim", '
        '"name": "common_packages", "start_time": '
        '"19:09:59.565546", "result": null, '
        '"duration": 412.241, "__run_num__": 0, '
        '"__sls__": "common.packages", '
        '"changes": {}, '
        '"__id__": "common_packages"}, '
        '"pkg_|-install_alcali_|-install_alcali_'
        '|-installed": {"comment": "All specified '
        'packages are already installed", "name": '
        '"curl", "start_time": "19:09:59.978059", '
        '"result": true, "duration": 34.526, '
        '"__run_num__": 1, "__sls__": "alcali", '
        '"changes": {}, "__id__": '
        '"install_alcali"}}, '
        '"retcode": 0, "success": true, '
        '"cmd": "_return", "_stamp": '
        '"2019-05-07T19:10:00.018461", '
        '"fun": "state.apply", "id": "master", '
        '"out": "highstate"}',
        alter_time="2019-05-07 19:10:00",
    )
    return ret


@pytest.fixture
def minion():
    minion = Minions.objects.create(
        minion_id="2e220fd40bc5", grain="thisisnotvalid", pillar="same"
    )
    return minion


@pytest.fixture
def dummy_jid():
    Jids.objects.create(
        jid="20190507190955945844",
        load='{"tgt_type": "glob", "jid": "20190507190955945844", '
        '"tgt": "master", "cmd": "publish", "ret": "", '
        '"user": "user1", "key": '
        '"N6TS4++wIWU/7jOQsSYO+E42VbFc/b/j3nD9GC24Oo08jF'
        '/bY49vgksEHUMmRNzLXTuzsdiYLwM=", "kwargs": {"token": '
        '"3ee91b11a58f8265c26db179dd61244b12db375d9855d4b2487033f7'
        '540710ca", "client": "local", "batch": null}, "fun": '
        '"alcali.pass_salt", "arg": []}',
    )


@pytest.fixture
def dummy_state():
    ret = SaltReturns.objects.create(
        fun="alcali.pass_salt",
        jid="20190507190955945844",
        return_field='{"oh": "no"}',
        id="master",
        success=1,
        full_ret='{"return": "noice"}',
        alter_time=time.strftime("%Y-%m-%d %H:%M:%S"),
    )
    return ret


@pytest.fixture
def minion_master():
    ret = Minions.objects.create(
        minion_id="master",
        grain='{"biosversion": "1.5.1", "kernel": "Linux", '
        '"domain": "", "uid": 0, "zmqversion": "4.2.1", '
        '"kernelrelease": "5.0.11-300.fc30.x86_64", '
        '"pythonpath": ['
        '"/usr/lib/python2.7/dist-packages/git/ext'
        '/gitdb", "/usr/bin", "/usr/lib/python2.7", '
        '"/usr/local/lib/python2.7/dist-packages", '
        '"/usr/lib/python2.7/dist-packages", '
        '"/usr/lib/python2.7/dist-packages/gitdb/ext'
        '/smmap"], "serialnumber": "HSZHST2", '
        '"pid": 59, "fqdns": [], "ip_interfaces": {'
        '"lo": ["127.0.0.1"], "eth0": ["172.29.0.4"]}, '
        '"groupname": "root", "fqdn_ip6": [], '
        '"mem_total": 15480, "saltversioninfo": [2019, '
        '2, 0, 0], "zfs_support": false, "SSDs": ['
        '"dm-1", "nvme0n1", "dm-2", "dm-0", "dm-3"], '
        '"mdadm": [], "id": "master", "manufacturer": '
        '"Dell Inc.", "osrelease": "9.8", "ps": "ps '
        '-efHww", "locale_info": {"timezone": "UTC", '
        '"detectedencoding": "UTF-8", '
        '"defaultlanguage": "en_US", "defaultencoding": '
        '"UTF-8"}, "fqdn": "4181631c3821", '
        '"lsb_distrib_os": "GNU/Linux", "ip_gw": true, '
        '"ip6_interfaces": {"lo": [], "eth0": []}, '
        '"num_cpus": 8, "hwaddr_interfaces": {"lo": '
        '"00:00:00:00:00:00", "eth0": '
        '"02:42:ac:1d:00:04"}, "osfullname": "Debian", '
        '"ip4_interfaces": {"lo": ["127.0.0.1"], '
        '"eth0": ["172.29.0.4"]}, "init": "unknown", '
        '"gid": 0, "master": "master", '
        '"virtual_subtype": "Docker", "dns": {"domain": '
        '"", "sortlist": [], "nameservers": ['
        '"127.0.0.11"], "ip4_nameservers": ['
        '"127.0.0.11"], "search": [], '
        '"ip6_nameservers": [], "options": ['
        '"ndots:0"]}, "ipv6": [], "cpu_flags": ["fpu", '
        '"pts", "hwp", "hwp_notify", "hwp_act_window", '
        '"hwp_epp", "flush_l1d"], "localhost": '
        '"4181631c3821", "ipv4": ["127.0.0.1", '
        '"172.29.0.4"], "username": "root", "fqdn_ip4": '
        '["172.29.0.4"], "shell": "/bin/sh", '
        '"nodename": "4181631c3821", "saltversion": '
        '"2019.2.0", "lsb_distrib_release": "9.8", '
        '"ip6_gw": false, "server_id": 685245236, '
        '"saltpath": '
        '"/usr/lib/python2.7/dist-packages/salt", '
        '"zfs_feature_flags": false, "osmajorrelease": '
        '9, "swap_total": 7807, "os_family": "Debian", '
        '"oscodename": "stretch", "osfinger": '
        '"Debian-9", "pythonversion": [2, 7, 13, '
        '"final", 0], "lsb_distrib_description": '
        '"Debian GNU/Linux 9.8 (stretch)", '
        '"kernelversion": "#1 SMP Thu May 2 14:11:38 '
        'UTC 2019", "ip4_gw": "172.29.0.1", "num_gpus": '
        '1, "virtual": "physical", "host": '
        '"4181631c3821", "disks": [], "cpu_model": '
        '"Intel(R) Core(TM) i7-8550U CPU @ 1.80GHz", '
        '"uuid": '
        '"4c4c4544-0053-5a10-8048-c8c04f535432", '
        '"biosreleasedate": "08/09/2018", '
        '"productname": "XPS 13 9370", "osarch": '
        '"amd64", "cpuarch": "x86_64", '
        '"lsb_distrib_codename": "stretch", '
        '"osrelease_info": [9, 8], "lsb_distrib_id": '
        '"Debian", "gpus": [{"model": "Device 5917", '
        '"vendor": "intel"}], "path": '
        '"/usr/local/bin:/usr/local/sbin:/usr/local/bin'
        ':/usr/sbin:/usr/bin:/sbin:/bin", "machine_id": '
        '"8b6f0bc66c81d2b9ef76a3fc9b52b452", '
        '"os": "Debian", "pythonexecutable": '
        '"/usr/bin/python2"}',
        pillar='{"foo": "bar", "list": ["first", "second", ' '"third"]}',
    )
    return ret


@pytest.fixture
def schedule(minion_master):
    ret = Schedule.objects.create(
        minion=minion_master,
        name="job2",
        job='{"master": {"job1": {"function": "test.ping", '
        '"name": "job1", "seconds": 3600, "enabled": '
        'false, "splay": null, "jid_include": true, '
        '"maxrunning": 1}}}',
    )


@pytest.fixture
def dummy_user():
    return User.objects.create_superuser(
        "dummy_user", "dummy_user@example.com", "dummy_userpassword"
    )


@pytest.fixture
def foo_conformity():
    return Conformity.objects.create(name="foo", function="alcali.pass_salt")


@pytest.fixture
def version_conformity():
    return Conformity.objects.create(
        name="version", function="cmd.run alcali --version"
    )


@pytest.fixture
def alcali_version_state():
    ret = SaltReturns.objects.create(
        fun="cmd.run",
        jid="20190507190955945844",
        return_field='{"oh": "no"}',
        id="master",
        success=1,
        alter_time=time.strftime("%Y-%m-%d %H:%M:%S"),
        full_ret='{"fun_args": ["alcali --version"], "return": "2019.2.2"}',
    )
    return ret


@pytest.fixture
def jobs_arguments():
    ret = SaltReturns.objects.create(
        fun="dummy.state",
        jid="20190507190955945844",
        return_field='{"oh": "no"}',
        id="master",
        success=1,
        alter_time=time.strftime("%Y-%m-%d %H:%M:%S"),
        full_ret='{"fun_args": ["foo", 1, "bar=baz"], "return": "2019.2.2"}',
    )
    return ret


@pytest.fixture
def highstate_failed():
    ret = SaltReturns.objects.create(
        fun="state.apply",
        jid="20190507190955945843",
        return_field='{"pkg_|-common_packages_'
        '|-common_packages_|-installed": {'
        '"comment": "The following packages '
        "would "
        'be installed/updated: vim", "name": '
        '"common_packages", "start_time": '
        '"19:09:59.565546", "result": null, '
        '"duration": 412.241, "__run_num__": '
        "0, "
        '"__sls__": "common.packages", '
        '"changes": '
        '{}, "__id__": "common_packages"}, '
        '"pkg_|-install_alcali_'
        "|-install_alcali_ "
        '|-installed": {"comment": "All '
        "specified "
        'packages are already installed", '
        '"name": '
        '"curl", "start_time": '
        '"19:09:59.978059", '
        '"result": true, "duration": 34.526, '
        '"__run_num__": 1, "__sls__": '
        '"alcali", '
        '"changes": {}, "__id__": '
        '"install_alcali"}}',
        id="2e220fd40bc5",
        success=1,
        full_ret='{"fun_args": ["test=True"], "jid": '
        '"20190507190955945843", "return": ["SLS Data failed to compile"], '
        '"retcode": 0, "success": false, '
        '"cmd": "_return", "_stamp": '
        '"2019-05-07T19:10:00.018461", '
        '"fun": "state.apply", "id": "master", '
        '"out": "highstate"}',
        alter_time="2019-05-07 19:10:00",
    )
    return ret



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("requirements/prod.txt", "r") as fh:
    requirements = fh.read().splitlines()

with open("requirements/dev.txt", "r") as fh:
    dev_requirements = fh.read().splitlines()

with open("VERSION", "r") as fh:
    version = fh.read()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="alcali",
    version=version,
    author="Matt Melquiond",
    author_email="matt.LLVW@gmail.com",
    description="Alcali",
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/latenighttales/alcali.git",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Systems Administration",
    ],
    install_requires=requirements,
    extras_require={"dev": dev_requirements},
    entry_points={"console_scripts": ["alcali = bin:manage"]},
)
